import shutil

import rasterio
import argparse
import os
from os.path import join
import traceback
import numpy as np

import frame_s2.postproc_helper as postproc_helper
from frame_s2.config.FRAME_settings import lv1_settings
from frame_s2.postproc_helper import  tif_update_metadata, dekadToStartDate, validate_inputs, previousDekad, Spectral_file, run_command, dateToDekad
from frame_s2.albedo import calc_and_mask_albedo


def parse_arguments():
    parser = argparse.ArgumentParser(prog='lv1_import_and_mask.py')

    parser.add_argument("--mask", type=str, help="Land/Sea MASK-IMG (byte, no extension, 0=Water, 1-255=Land",
                        required=True,
                        default=lv1_settings['LS_MASK_FILE'])
    parser.add_argument("--dekad_start_date", type=str, help="dekad_start_date in format YYYYMMDD",
                         required=True)
    parser.add_argument("--root_dir", type=str, help="root dir for lv1 processing",
                        required=False)
    parser.add_argument("--normalisation", type=str, help="Divide SUM[Wi x Ri] by SUM[Wi]  [0/1=def]",
                        default=1)
    return parser.parse_args()


def import_modis(preprocessing_dir, outDir, dekad, dekad_date):
    print("Converting img to tif ....")

    '''unzipping and converting to tif'''
    #   unzip
    # zipfile = join(preprocessing_dir, dekad_date, 'mt%s.zip') % dekad
    # cmd = 'unzip %s %s' % (zipfile, outDir)
    # stdout, stderr, err_nr = run_command(cmd)
    # if err_nr:
    #     print(traceback.print_exc())
    #     print(stdout)
    #     print(stderr)
    #     raise Exception("Unable to unzip %s" % zipfile)
    #   convert to tif
    for band, bandproperties in lv1_settings['modis_bands'].items():
        img_fname = join(preprocessing_dir, dekad_date, 'mt%s%s.img') % (dekad, band)
        tif_fname = join(outDir, dekad_date, 'mt%s%s.tif') % (dekad, band)
        tif_dir = os.path.dirname(tif_fname)
        if not os.path.exists(tif_dir):
            try:
                os.mkdir(tif_dir)
            except:
                raise Exception("Unable to create %s" % tif_dir)
        cmd = "gdal_translate -co 'BLOCKXSIZE=256' -co 'BLOCKYSIZE=256' -co 'TILED=YES' -co 'COMPRESS=LZW' %s %s" % (
        img_fname, tif_fname)
        #cmd = "echo No gdal_translate in testing mode"
        stdout, stderr, err_nr = run_command(cmd)
        if err_nr:
            print(traceback.print_exc())
            print(stdout)
            print(stderr)
            raise Exception("Converting img to tif failed for file %s" % img_fname)

    print("  Finished successfully with writing tifs to folder %s" % tif_dir)
    return tif_dir

def mask_ndvi(in_ndvi_file, out_ndvi_file, mask_file):
    with rasterio.open(in_ndvi_file, 'r', sharing=False) as src:
        profile = src.profile

    # set output
    profile.update(
        dtype=rasterio.uint8,
        count=1,
        compress='lzw',
        BIGTIFF='YES')

    with rasterio.open(out_ndvi_file, 'w+', **profile, sharing=False) as dst:
        with rasterio.open(in_ndvi_file, 'r', sharing=False) as src:
            with rasterio.open(mask_file, 'r', sharing=False) as msk:
                print("Masking NDVI ....")
                for ji, window in dst.block_windows(1):
                    #print ('ji : %s, %s' % (ji[0], ji[1]))
                    ndvi_in = src.read(1, window=window).astype(np.ubyte)
                    msk_array = msk.read(1, window=window).astype(np.ubyte)
                    ndvi_out = postproc_helper.mask_data(ndvi_in, msk_array, pixel_convert_lst=[(254, 251)], new_mask_val=254 )
                    dst.write(ndvi_out, 1, window=window)

    #update tif metadata
    ndvi_out = tif_update_metadata(out_ndvi_file,
                                    scale=lv1_settings['NDV_SCALE'],
                                    offset=lv1_settings['NDV_OFFSET'],
                                    description=lv1_settings['NDV_DESCRIPTION'],
                                    units=lv1_settings['NDV_UNITS'],
                                    flags=lv1_settings['NDV_FLAGS'],
                                    nodata=lv1_settings['NDV_NODATA']
                                    )
    print("  Finished successfully with writing %s" % ndvi_out)


def start_import_and_mask_lv1( root_dir, mask_file, in_folder, normalisation, clean_tmp=True):
    #in_folder will be the subdir in format YYYYMMDD where all modis files are located
    preprocessing_dir = os.path.dirname(in_folder.rstrip('/'))  #yyyymmdd should be skipped here
    dekad_date = os.path.basename(in_folder.rstrip('/'))
    dekad = dateToDekad(dekad_date)
    #set folders if they do not exist
    #preprocessing_dir = join(root_dir, 'syn/mod')
    alb_dir = join(root_dir, 'alb', 'msk')
    ndv_msk_dir = join(root_dir, 'ndv', 'msk')
    tmp_dir = join(root_dir, 'tmp')
    if not os.path.exists(tmp_dir):
         os.mkdir(tmp_dir)

    # if not os.path.exists(alb_dir):
    #     os.mkdir(alb_dir)

    # if not os.path.exists(ndv_dir):
    #     os.mkdir(ndv_dir)

    ##0) checks
    if not os.path.exists(mask_file):
        raise Exception("Mask-file does not exsist : %s" % mask_file)

    ##1) import
    print ('Importing...')
    import_dir = import_modis(preprocessing_dir, tmp_dir, dekad, dekad_date)
    #import_dir = '/data/FRAME/lv1/tmp/20210821/'

    ##2) NDVI
    in_ndvi_file = join(import_dir, 'mt%s5.tif') % dekad
    out_ndvi_file = join(ndv_msk_dir, 'L1_NDVI_%s.tif') % dekad
    mask_ndvi(in_ndvi_file, out_ndvi_file, mask_file)


    # 3) ALBEDO
    mask_file = out_ndvi_file
    albedo_file = calc_and_mask_albedo(import_dir, alb_dir, mask_file, dekad, NORM=normalisation)

    #4) Clean tmp inputs
    if clean_tmp:
        print ("Cleaning %s" % import_dir)
        shutil.rmtree(import_dir)
    print('Byebye')


def main():
    args = parse_arguments()
    start_import_and_mask_lv1(args.root_dir, args.mask, args.dekad_start_date, args.normalisation)

if __name__ == "__main__":
    main()