# -*- coding: utf-8 -*-
from math import exp, floor
import argparse
import rasterio
import numpy as np
import numpy.ma as ma
import os
from os.path import join

from frame_s2.config.FRAME_settings import meteo_settings_dict, lv2_settings
from frame_s2.postproc_helper import rescale, tif_update_metadata, create_quicklook



#def Plut(short Otyp, double ARint, double ARslo, double PhotEff, double CO2act, double Tc1, double Tc2, double TcDelt, float *LUTeff, float *LUTco2, float *LUTar):
def createLUT(infile):
    LUT={}
    RUElut = np.full(256, -1.0)
    with open(infile, "r") as f:
        lines = f.readlines()
    for line in lines:
        #remove empty lines
        if len(line.strip()) == 0:
            continue
        # skip the lines starting with #
        if line.strip()[0] == "#":
            continue

        # FAPAR, RUE, classification
        lst = line.split(',')
        idx = int(lst[0].strip())
        val = float(lst[1].strip())
        RUElut[idx] = val

    return RUElut

def calc_npp(fapar, pmax, LU_class, RUElut):

    #validation
    if np.any(np.where(LU_class < 0.0)):
        raise Exception("LU-class contains values less than 0")

    ##CREATE STATS
    #currently only fapar=255 is masked(nodata), also mask 251 now and count the missing values
    count_missing_in_fapar = np.count_nonzero(fapar == 251)
    fapar = ma.masked_greater(fapar, 250)

    #stats for pmax
    count_no_meteo = ma.count_masked(pmax)

    #start filling the Array with RUE parameters using the LUT table
    rue = RUElut[LU_class]
    rue_masked = ma.masked_equal(rue, 0.0)
    #pmax was not rescaled for this, so we are working with it's DN here
    npp = pmax * rue_masked * fapar

    #Keep in range [0 - 32767]
    npp = ma.where(npp > 32767, 32767, npp)
    npp = ma.where(npp < 0, 0, npp)
    #convert to int
    npp = np.floor(npp + 0.5).astype(np.int16)  # use rounding half up method

    #creating stats

    #count_missing_in_fapar = ma.count_masked(fapar)
    count_water_in_LUT = np.count_nonzero(rue == 0.0)

    count_ok = ma.count(npp)
    count_total = ma.size(npp)
    #count_snow_ice = np.count_nonzero(fapar=)
    return npp, np.array([count_total, count_ok, count_no_meteo, count_missing_in_fapar, count_water_in_LUT], np.uint32)

def parse_arguments():
    #--start_dekad 2113 --end_dekad 2113 --in_fapar_dir /data/FRAME/pact --in_pmax_dir /data/FRAME/pact --in_classification /data/FRAME/pact/LU2015_SRI.tif --in_RUE LU_RUE.txt --out_npp_dir /data/FRAME/pact --roi SRI
    #--start_dekad 2118 --end_dekad 2118 --in_fapar_dir /data/FRAME/pact/tests_herman/fpr --in_pmax_dir /data/FRAME/pact/tests_herman/meteo --in_classification /data/FRAME/pact/tests_herman/LU2015.tif --in_RUE LU_RUE.txt --out_npp_dir /data/FRAME/pact/tests_herman/ --roi SRI
    parser = argparse.ArgumentParser(prog='pmax.py',description="Create IMGs with MAX. DMP/NPP from LoRes METEO-DATA")

    parser.add_argument("--in_fapar", type=str, help="FAPAR image",
                        required=True)
    parser.add_argument("--in_pmax", type=str, help="pmax image",
                        required=True)
    parser.add_argument("--in_classification", type=str, help="classification image",
                        required=True)
    parser.add_argument("--in_RUE", type=str, help="LU-class, RUE, text file",
                        required=True)
    parser.add_argument("--out_npp", type=str, help="output npp file",
                        required=True)
    parser.add_argument("--out_nodata", type=int, help="nodata value for output file",
                        required=False,
                        default=-9999)
    parser.add_argument("--roi", type=str, help="SRI|AFR|SSB",
                        required=True)
    parser.add_argument("--dekad", type=str, help="dekad to process",
                        required=True)
    parser.add_argument("--level", type=int, help="1:Lv1, 2:Lv2",
                        required=False,
                        default=2)
    return parser.parse_args()

def unscale_inputs(fapar_in, pmax_in):
    fapar_unscaled = rescale(fapar_in,
                          0,
                          1,
                          max_value_in=200,
                          min_value_in=0,
                          max_value_out=250,
                          min_value_out=0,
                          offset_in=lv2_settings['FPR_OFFSET'],
                          scale_in=lv2_settings['FPR_SCALE'],
                          out_data_type=np.double)
    fapar_masked = ma.masked_equal(fapar_unscaled, 255)
    #we do not rescale the pmax values because the end result has the same scale and offset
    pmax_unscaled = rescale(pmax_in,
                          0,
                          1,
                          max_value_in=32767,
                          min_value_in=0,
                          max_value_out=32767,
                          min_value_out=0,
                          offset_in=0,
                          scale_in=1,
                          out_data_type=np.double)
    pmax_masked = ma.masked_equal(pmax_unscaled, -9999)

    return fapar_masked, pmax_masked

def validate_inputs(in_img_list):
    bbox_list=[]
    shape_list=[]
    for img in in_img_list:
        with rasterio.open(img) as src:
            shape_list.append(src.shape)
            bbox_list.append(src.bounds)

            #validating blocksize
            blockxsize = int(src.profile["blockxsize"])
            blockysize = int(src.profile["blockysize"])
            if blockxsize != meteo_settings_dict["input_blockXsize"] or blockysize != meteo_settings_dict["input_blockYsize"]:
                raise Exception ("Input blocksize %s,%s does not match the expected ones %s,%s"
                                 % (blockxsize,
                                    blockysize,
                                    meteo_settings_dict["input_blockXsize"],
                                    meteo_settings_dict["input_blockYsize"]))

    #validate if all items in the list are the same, by using set
    shape_set = set(shape_list)
    bbox_set = set(bbox_list)
    if len(shape_set)!= 1: raise Exception ('Shapes of the in images has differences : %s for files %s' % (shape_list, in_img_list))

    #check if all coordinates of the bbox are equal, keep in mind that some rounding errors can be there
    left=[];right=[];top=[];bottom=[]
    digits=7    #7 digits is less than a meter (for 100m data)
    for bbox in bbox_list:
        left.append(round(bbox[0], digits))
        bottom.append(round(bbox[1], digits))
        right.append(round(bbox[2], digits))
        top.append(round(bbox[3], digits))
    left_set=set(left)
    bottom_set=set(bottom)
    right_set=set(right)
    top_set=set(top)

    #if len(bbox_set) != 1: raise Exception('Shapes of the in images has differences : %s' % bbox_list)
    if len(left_set) != 1 or len(bottom_set)!= 1 or len(right_set)!=1 or len(top_set)!=1 :
        raise Exception('Shapes of the in images has differences : %s' % bbox_list)


    total_number_of_blocks = round(int(shape_list[0][0])/blockxsize * int(shape_list[0][1])/blockysize)
    if total_number_of_blocks == 1: print("-Warn- Total number of blocks is 1! For big images this can couse memory" \
                                               " issues. To solve this, provide is tiff with blocks of eg 256:256")
    return total_number_of_blocks

def main():
    args = parse_arguments()
    start_pact(args.in_fapar, args.in_pmax, args.in_classification, args.in_RUE, args.out_npp, args.dekad, args.roi, args.out_nodata, args.level)

def start_pact(in_fapar_dir, in_pmax_dir, in_classification, in_RUE, out_npp_dir, dekad, roi, out_nodata=-9999, level=2):
    #set gdal cache parameter on the workers to reduce memory usage.
    # rasterio uses gdal in the background so limit the GDAL_CACHE
    # this parameter is in bytes. 1GB wil result in 1,2GB of memory usage(tested on local machine)
    os.environ['GDAL_CACHEMAX'] = meteo_settings_dict['GDAL_CACHEMAX']

    stats_total = np.array([0.0, 0.0, 0.0, 0.0, 0.0], dtype=np.float32)     #init, fill with empty stats

    #set filenames according to level and roi
    if not level in [1, 2]: raise Exception("Level is wrong! is %s, but should be 1 or 2" % level)
    if level == 2:
        if roi == 'AFR':
            in_pmax = join(in_pmax_dir, 'L2_AFR_NPM_%s.tif' ) % ( dekad)
            in_fapar = join(in_fapar_dir, 'L2_FPR_%s.tif' ) % (dekad)
            out_npp = join(out_npp_dir, 'L2_NPP_%s.tif') % ( dekad)
        else:
            in_pmax = join(in_pmax_dir, 'L2_%s_NPM_%s.tif') % (roi, dekad)
            in_fapar = join(in_fapar_dir, 'L2_%s_FPR_%s.tif') % (roi, dekad)
            out_npp = join(out_npp_dir, 'L2_%s_NPP_%s.tif') % (roi, dekad)
    else:
        #level=1
        in_pmax = join(in_pmax_dir, 'L1_%s_NPM_%s.tif') % (roi, dekad)
        in_fapar = join(in_fapar_dir, 'L1_FPR_%s.tif') % (dekad)
        out_npp = join(out_npp_dir, 'L1_NPP_%s.tif') % (dekad)


    #validation
    total_number_of_blocks = validate_inputs([in_fapar, in_pmax, in_classification])
    #create LUT table for classification map
    RUElut = createLUT(in_RUE)
    with rasterio.open(in_fapar, 'r', sharing=False) as src_sr:
        sr_profile = src_sr.profile
        #set output
        profile = sr_profile
        profile.update(
            dtype=rasterio.int16,
            count=1,
            compress='lzw',
            BIGTIFF='YES')

    #open all files for blockprocessing
    # read block per block
    print ("Start processing %s" % out_npp)
    try:
        if os.path.exists(out_npp):
            os.remove(out_npp)
    except:
        print ("-W- Unable to delete %s" % out_npp)
    with rasterio.open(out_npp, 'w', sharing=False, **profile) as dst:
        with rasterio.open(in_fapar, 'r', sharing=False) as src_fapar:
            with rasterio.open(in_pmax, 'r', sharing=False) as src_pmax:
                try:
                    used_meteo_lst = src_pmax.tags()['Used_inputs']
                except:
                    used_meteo_lst = 'No information available'
                with rasterio.open(in_classification, 'r', sharing=False) as src_classification:
                    for ji, window in src_fapar.block_windows(1):
                        fapar = src_fapar.read(1, window=window)
                        pmax = src_pmax.read(1, window=window)
                        classification = src_classification.read(1, window=window)

                        #unscale the data for calculations : Convert to Physical units: kJ/m²/day or Decigrades Celsius (NEO: V1608)
                        fapar_unscaled, pmax_unscaled = unscale_inputs(fapar, pmax)

                        #sr_unscaled as a different mask then tmin and tmax, calculate the overlapping mask of all products and add it to the 3 arrays
                        #overlapping_mask = tmin_unscaled.mask * tmax_unscaled.mask * sr_unscaled.mask
                        npp, stats = calc_npp(fapar_unscaled, pmax_unscaled, classification, RUElut)
                        npp = npp.filled(fill_value=-9999)
                        #write result to disk
                        dst.write(npp.astype(rasterio.int16), 1, window=window)
                        stats_total+=stats

        dst.update_tags(Used_inputs=used_meteo_lst)
        dst.nodata = out_nodata

    #setting scale and offset to tif
    out_npp = tif_update_metadata(out_npp, scale=meteo_settings_dict['NPP_SCALE'], offset=meteo_settings_dict['NPP_OFFSET'],
                                  description=meteo_settings_dict['NPP_DESCRIPTION'], units=meteo_settings_dict['NPP_UNITS'])

    #create quicklook
    ql_width = lv2_settings['ROIS'][roi]['ql_width']
    ql_height = lv2_settings['ROIS'][roi]['ql_height']
    create_quicklook(out_npp, 'npp.txt', scale='0 32767 1 255', height=ql_height, width=ql_width)
    print("Finished successfully with writing %s" % out_npp)

    #count_total, count_ok, count_no_meteo, count_missing_in_fapar, count_water_in_LUT
    total = stats_total[0]/100.0  #stats in percent
    stat = stats_total[0] / total
    print("Total : %.2f" % stat)
    stat = stats_total[1] / total
    print("Normally treated : %.2f" % stat)
    stat = stats_total[2]
    print("No meteo/pmax: %.2f" % stat)
    stat = stats_total[3] / total
    print("No fapar: %.2f" % stat)
    stat = stats_total[4] / total
    print("Water: %.2f" % stat)


if __name__ == "__main__":
    main()