# -*- coding: utf-8 -*-
#--in_NDVI_lst /data/FRAME/PHE/inputs_test/NDV.lst --out_folder /data/FRAME/PHE/output_test/ --roi SRI --year 2019
import argparse
import rasterio
import numpy as np
import numpy.ma as ma
from os.path import join
import os
import sys
from ctypes import  c_double, c_short, c_float, CDLL, byref, Structure, POINTER, c_ubyte, c_int,c_void_p, c_int64, cast, c_byte
import ctypes

from numpy.ctypeslib import ndpointer

from frame_s2.config.FRAME_settings import meteo_settings_dict, lv2_settings
from frame_s2.config.Pheno_settings import pheno_settings
from frame_s2.config.pheno_kk0_colormap import kk0_colormap
from frame_s2.postproc_helper import tif_update_metadata, validate_inputs, check_chronological_order_input_lst

##### CONSTANTS ############
NUMBER_IN_FILES  = 108   # NvI : Nr. of IN-IMGs: 108 = 3 years * 36 dekads
NUMBER_OUT_FILES = 22    # NvO :Nr. of maximal possible OUT-IMGs
OUT_SUFFIX=["Ya0", "Yr0", "KK0",                                       # 00-02: Annual values (independent of seasons) : Y-mean, Y-range, Classification
          "s1",  "m1",  "e1",  "L1",  "s2",  "m2",  "e2",  "L2",    # 03-10: Dekads of SOS, MOS, EOS & Season length,  for 2 seasons
         "Ys1", "Ym1", "Ye1", "Ya1", "Ys2", "Ym2", "Ye2", "Ya2",    # 11-18: Values at SOS, MOS, EOS & Season average, for 2 seasons
         "LT0", "LR0", "SI0"]


class HdrInfo(ctypes.Structure):
    _fields_ = [
        ("Vlo", ctypes.c_double),
        ("Vhi", ctypes.c_double),
        ("Vint", ctypes.c_double),
        ("Vslo", ctypes.c_double)
    ]

so_file='pheno.so'
if not os.path.exists(so_file):
    so_file = 'executables/pheno.so'
lib = CDLL(so_file)
#settings the global Settings structure, these settings are hard coded in C
lib_Set_settings = lib.Set_settings
lib_Set_settings.argtypes = None
lib_Set_settings.restype = None

#settings the global Settings structure, these settings are hard coded in C
lib_Allocate_mem = lib.Allocate_mem
lib_Allocate_mem.argtypes = None
lib_Allocate_mem.restype = None

#loop(const unsigned char *arr_in, unsigned char *arr_pix, unsigned char *arr_out, const int n_in, const int n_out, const int rows, const int columns);

lib_loop = lib.loop
lib_loop.argtypes = [
    np.ctypeslib.ndpointer(dtype=np.ubyte, ndim=3, flags='C_CONTIGUOUS'),
    np.ctypeslib.ndpointer(dtype=np.ubyte, ndim=1, flags='C_CONTIGUOUS'),
    np.ctypeslib.ndpointer(dtype=np.ubyte, ndim=3, flags='C_CONTIGUOUS'),
    c_short, c_short, c_short, c_short]
lib_loop.restype = None

lib_Set_parameters = lib.Set_Parameters
lib_Set_parameters.argtypes = [c_ubyte, c_ubyte, c_double, c_double, c_double]
lib_Set_parameters.restype = None

lib_Set_OutHeaders = lib.Set_OutHeaders
lib_Set_OutHeaders.argtypes = [HdrInfo, c_short]
lib_Set_OutHeaders.restype = None


# class SEASON(ctypes.Structure):
#     #Additional parameters: Inputs & Derived
#     _fields_ = [
#         ("SOS", ctypes.c_short), ("MOS", ctypes.c_short), ("EOS", ctypes.c_short), ("L", ctypes.c_short), ("Keep", ctypes.c_short),
#         ("Ysos", ctypes.c_double), ("Ymos", ctypes.c_double), ("Yeos", ctypes.c_double),("Area", ctypes.c_double)
#         ]

def Set_Global_C_Vars():
    #SETTING NDVI PARAMETERS
    print ('---SETTING NDVI IN PARAMETERS----')
    Vlo=0;Vhi=250;Vint=-0.08;Vslo=0.004;MaxMisPerc=15;
    lib_Set_parameters(int(Vlo), int(Vhi), float(Vint), float(Vslo), float(MaxMisPerc))

    #SETTING OUTPUT HEADERS C-var
    print('---SETTING OUTPUT PARAMETERS----')
    num = int(0)
    for key, setting in pheno_settings['out'].items():
        # self.HDR_ARRAY[num].Vindex = key
        hdr = HdrInfo(float(setting['Vlo'] or 0), float(setting['Vhi'] or 255), float(setting["Vint"] or 0), float(setting["Vslo"] or 1))
        #hdr = HdrInfo(8.1*num, 1.0*num, 3.0*num, 4.0*num)
        lib_Set_OutHeaders(hdr, key)
        num += 1

    #SETTING other settings
    lib_Set_settings()
    lib_Allocate_mem()
    return
    #simple_headers_out = (POINTER(HdrInfo) * NUMBER_OUT_FILES)()
    # #CYC
    # cyc = (POINTER(SEASON) * 20)
    # CYC = cast(cyc, POINTER(SEASON))
    #
    # #SSN
    # ssn = (POINTER(SEASON) * 2)
    # SSN = cast(ssn, POINTER(SEASON))

    # return simple_headers_out, CYC, SSN


#plut_c.restype = ndpointer(dtype=c_double,shape=(n,)) #returning numpy array

# Python representation of the C struct re_val
# class ReVal(Structure):
#     _fields_ = [("LUTeff", POINTER(c_double)),("LUTco2", POINTER(c_double)),("LUTar", POINTER(c_double))]
# plut_c.restype = ReVal
#calling C-program stuff


def parse_arguments():
    parser = argparse.ArgumentParser(prog='pheno.py',description="Create phenology output")

    parser.add_argument("--in_NDVI_lst", type=str, help="a file with 108 NDVI paths, sorted(increasing in time)!!",
                        required=True)
    parser.add_argument("--out_folder", type=str, help="output nok file",
                        required=True)
    parser.add_argument("--roi", type=str, help="roi=AFR|SRI|SSB",
                        required=True)
    parser.add_argument("--year", type=str, help="year in format YYYY",
                        required=True)
    return parser.parse_args()

def calc_pheno_by_matrix(in_data_cube):
    #calculation phenology using matrixes

    # initialize C-arrays
    n_in, rows, columns = in_data_cube.shape
    in_data_cube=in_data_cube.astype(np.uint8)
    if n_in !=NUMBER_IN_FILES:
        raise Exception("nr of in files is not 108")
    result_cube = np.full([ NUMBER_OUT_FILES, rows, columns], 255, np.uint8)
    arr_pix = np.empty(NUMBER_IN_FILES, np.uint8)
    lib_loop(in_data_cube, arr_pix, result_cube, NUMBER_IN_FILES, NUMBER_OUT_FILES, rows, columns )

    # #TODO
    # ''' """""""""""""""""""
    # if (S.o1 > 0)
    #     {
    #     for (i= 7;i <= 10;i++) {foM[i] = 0;}
    #     for (i=15;i <= 18;i++) {foM[i] = 0;}
    #     for (i=19;i <= 20;i++) {foM[i] = 0;}
    #     }'''
    return result_cube



def main():
    args = parse_arguments()
    #calc_pheno_by_matrix(np.zeros((2,2,2),np.ubyte))
    start_pheno(args.in_NDVI_lst, args.out_folder, args.roi, args.year)

def start_pheno(in_NDVI_lst_file, out_folder, roi, year):
    os.environ['GDAL_CACHEMAX'] = meteo_settings_dict['GDAL_CACHEMAX']


    #handling OUT's
    #for AFR, the roi abbreviation is not in the filenames, for the other roi's it is.
    if roi == 'AFR':
        roi_prefix = 'L2_PHE_%s_' % year
    elif roi == 'LV1':
        roi_prefix = 'L1_PHE_%s_' % year
    else:
        roi_prefix = 'L2_%s_PHE_%s_' % (roi, year)

    #roi_prefix = 'ndv_herman_nrt_my_lta_%s_' % year

    pheno_out_files_index_lst, pheno_out_files_name_lst = get_pheno_out_files_from_settings() #get a list of outfiles to be written, remember : all out parameters are processed!
    pheno_full_path_lst = list(map(lambda pheno_file: join(out_folder, roi_prefix + pheno_file + '.tif'), pheno_out_files_name_lst )) #create a list offull paths
    nr_out_files = len(pheno_full_path_lst)

    #handling IN's
    check_chronological_order_input_lst(in_NDVI_lst_file, 108)
    #create a list of NDVI input files
    with open(in_NDVI_lst_file,'r') as f:
        in_NDVI_lst = f.readlines()
    in_NDVI_lst = list(map(lambda ndvi: ndvi.strip(), in_NDVI_lst)) #remove line breaks from list
    #in_NDVI_lst.sort() ##"NEVER EVER SORT! sorting will reorder files based on the folder!!!
    print (in_NDVI_lst)
    print ("Number of input files : %s" % len(in_NDVI_lst))
    print ("Start processing...")


    #validation
    total_number_of_blocks = validate_inputs(in_NDVI_lst)

    # Creating profile with the first image from the list
    with rasterio.open(in_NDVI_lst[0], 'r', sharing=False) as src_sr:
        sr_profile = src_sr.profile

        #set output
        profile = sr_profile
        profile.update(
            dtype=rasterio.uint8,  #the same for all pheno output
            count=1,
            compress='lzw',
            BIGTIFF='YES')


    #setting global C-parameters
    Set_Global_C_Vars()

    #open outfiles and put the to the outfile lst:
    # writing the output
    pheno_out_files = pheno_settings['out']
    dst_lst=[]
    for index, pheno_file_settings in pheno_out_files.items():
        out_name = pheno_file_settings['name']
        out_fname = join(out_folder, roi_prefix + out_name + '.tif')
        dst = rasterio.open(out_fname, 'w+', **profile, sharing=False)
        if out_name == "KK0":
            dst.write_colormap(1, kk0_colormap)
            #write legend to the metadata
            dst.update_tags(_000='Ns=0/2 Mu=0/4 Range=0/4', _001='Ns=0/2 Mu=0/4 Range=1/4', _002='Ns=0/2 Mu=0/4 Range=2/4', _003='Ns=0/2 Mu=0/4 Range=3/4',
                            _004='Ns=0/2 Mu=0/4 Range=4/4', _010='Ns=0/2 Mu=1/4 Range=0/4', _011='Ns=0/2 Mu=1/4 Range=1/4', _012='Ns=0/2 Mu=1/4 Range=2/4',
                            _013='Ns=0/2 Mu=1/4 Range=3/4', _014='Ns=0/2 Mu=1/4 Range=4/4', _020='Ns=0/2 Mu=2/4 Range=0/4', _021='Ns=0/2 Mu=2/4 Range=1/4',
                            _022='Ns=0/2 Mu=2/4 Range=2/4', _023='Ns=0/2 Mu=2/4 Range=3/4', _024='Ns=0/2 Mu=2/4 Range=4/4', _030='Ns=0/2 Mu=3/4 Range=0/4',
                            _031='Ns=0/2 Mu=3/4 Range=1/4', _032='Ns=0/2 Mu=3/4 Range=2/4', _033='Ns=0/2 Mu=3/4 Range=3/4', _034='Ns=0/2 Mu=3/4 Range=4/4',
                            _040='Ns=0/2 Mu=4/4 Range=0/4', _041='Ns=0/2 Mu=4/4 Range=1/4', _042='Ns=0/2 Mu=4/4 Range=2/4', _043='Ns=0/2 Mu=4/4 Range=3/4',
                            _044='Ns=0/2 Mu=4/4 Range=4/4', _100='Ns=1/2 Mu=0/4 Range=0/4', _101='Ns=1/2 Mu=0/4 Range=1/4', _102='Ns=1/2 Mu=0/4 Range=2/4',
                            _103='Ns=1/2 Mu=0/4 Range=3/4', _104='Ns=1/2 Mu=0/4 Range=4/4', _110='Ns=1/2 Mu=1/4 Range=0/4', _111='Ns=1/2 Mu=1/4 Range=1/4',
                            _112='Ns=1/2 Mu=1/4 Range=2/4', _113='Ns=1/2 Mu=1/4 Range=3/4', _114='Ns=1/2 Mu=1/4 Range=4/4', _120='Ns=1/2 Mu=2/4 Range=0/4',
                            _121='Ns=1/2 Mu=2/4 Range=1/4', _122='Ns=1/2 Mu=2/4 Range=2/4', _123='Ns=1/2 Mu=2/4 Range=3/4', _124='Ns=1/2 Mu=2/4 Range=4/4',
                            _130='Ns=1/2 Mu=3/4 Range=0/4', _131='Ns=1/2 Mu=3/4 Range=1/4', _132='Ns=1/2 Mu=3/4 Range=2/4', _133='Ns=1/2 Mu=3/4 Range=3/4',
                            _134='Ns=1/2 Mu=3/4 Range=4/4', _140='Ns=1/2 Mu=4/4 Range=0/4', _141='Ns=1/2 Mu=4/4 Range=1/4', _142='Ns=1/2 Mu=4/4 Range=2/4',
                            _143='Ns=1/2 Mu=4/4 Range=3/4', _144='Ns=1/2 Mu=4/4 Range=4/4', _200='Ns=2/2 Mu=0/4 Range=0/4', _201='Ns=2/2 Mu=0/4 Range=1/4',
                            _202='Ns=2/2 Mu=0/4 Range=2/4', _203='Ns=2/2 Mu=0/4 Range=3/4', _204='Ns=2/2 Mu=0/4 Range=4/4', _210='Ns=2/2 Mu=1/4 Range=0/4',
                            _211='Ns=2/2 Mu=1/4 Range=1/4', _212='Ns=2/2 Mu=1/4 Range=2/4', _213='Ns=2/2 Mu=1/4 Range=3/4', _214='Ns=2/2 Mu=1/4 Range=4/4',
                            _220='Ns=2/2 Mu=2/4 Range=0/4', _221='Ns=2/2 Mu=2/4 Range=1/4', _222='Ns=2/2 Mu=2/4 Range=2/4', _223='Ns=2/2 Mu=2/4 Range=3/4',
                            _224='Ns=2/2 Mu=2/4 Range=4/4', _230='Ns=2/2 Mu=3/4 Range=0/4', _231='Ns=2/2 Mu=3/4 Range=1/4', _232='Ns=2/2 Mu=3/4 Range=2/4',
                            _233='Ns=2/2 Mu=3/4 Range=3/4', _234='Ns=2/2 Mu=3/4 Range=4/4', _240='Ns=2/2 Mu=4/4 Range=0/4', _241='Ns=2/2 Mu=4/4 Range=1/4',
                            _242='Ns=2/2 Mu=4/4 Range=2/4', _243='Ns=2/2 Mu=4/4 Range=3/4', _244='Ns=2/2 Mu=4/4 Range=4/4', _252='Processing error',
                            _253='Insufficient inputs', _254='Masked', _255='Water')

        dst_lst.append(dst)

    #Creating datacube_in with all NDVI files per tif block and calculate pheno
    print ("Building datacube...")
    with rasterio.open(in_NDVI_lst[0], 'r', sharing=False) as src0:
        for ji, window in src0.block_windows(1):
            # init input data_cube
            in_data_cube = np.full( (len(in_NDVI_lst), window.height, window.width), 0, np.uint8)

            #reading first input image
            ndv = src0.read(1, window=window).astype(np.uint8)
            in_data_cube[0, ...] = ndv

            # reading all other inputs and fill up in_data_cube
            for index, NDVI_file in enumerate(in_NDVI_lst[1:], start=1):
                #print ("Handling infile %s" % NDVI_file)
                src = rasterio.open(NDVI_file, 'r', sharing=False)
                ndv = src.read(1, window=window).astype(np.uint8)
                src.close()
                in_data_cube[index, ...] = ndv
                ndv = None

            # Start processing
            pheno = calc_pheno_by_matrix(in_data_cube)
            mask_array = None
            data_cube = None



            #writing blocks to files
            for index, dst in enumerate(dst_lst):
                dst.write(pheno[index].astype('uint8'), 1, window=window)

    # closting outfile
    for index, dst in enumerate(dst_lst):
        dst.close()
        out_name = os.path.basename(dst.name[:-4]).split('_')[-1]
        #get offset/scale from json and put this to the tif metadata
        for index, pheno_file_settings in  pheno_settings['out'].items():
            if pheno_file_settings['name'] == out_name:
                offset = pheno_file_settings["Vint"]
                scale = pheno_file_settings["Vslo"]
                units = pheno_file_settings["unit"]
                description = pheno_file_settings["description"]
                break

        out_name = tif_update_metadata(dst.name, scale=(scale or 1), offset=(offset or 0), description=description, units=units )
    print("Finished successfully with writing %s" % out_folder)

def get_pheno_out_files_from_settings():
    #this funtion will return an index list of all files that needs to be written to disk
    out_dict = pheno_settings['out']
    out_files_index_lst=[]  #result index list, containing all indexes for which we need an output file
    out_files_names_lst = []

    for index, value in out_dict.items():
        if value['generate_output']:
            fname = value['name']
            out_files_index_lst.append(index)
            out_files_names_lst.append(fname)
    return out_files_index_lst, out_files_names_lst

if __name__ == "__main__":
    main()