import rasterio
import argparse
import os
import datetime
from os.path import join

import numpy.ma as ma
import numpy as np

from frame_s2.config.FRAME_settings import lv2_settings
from frame_s2.postproc_helper import  tif_update_metadata, validate_inputs, previousDekad, create_quicklook
from frame_s2.fapar import start_fapar
from frame_s2.qua_adapt import start_qua_adapt


def parse_arguments():
    parser = argparse.ArgumentParser(prog='postprocess.py')

    parser.add_argument("--roi", type=str, help="roi",
                        required=True)
    parser.add_argument("--dekad", type=str, help="dekad",
                         required=True)
    parser.add_argument("--input_dir", type=str, help="directory of the mosaics, filename should match 'composite_[QUA|ALB|NDV]_[dekad].tif",
                        required=False)
    parser.add_argument("--output_dir", type=str, help="Output root directory",
                        required=True)
    parser.add_argument("--nrOfHistoricalDekads", type=str, help="The number of historical dekads for a given dekad, default is 6 for nrt",
                        required=True)
    parser.add_argument("--prev_qua_dek_img", type=str, help="Used in QUA_adapt as source for the current QUA image, if None, this value will be calculate the previous one. If that one doesn't exist, use flag values.",
                        default=None)
    return parser.parse_args()

def rescale_S2_Proba(in_arr, offset_out, scale_out, max_value_in=250, min_value_in=0, max_value_out=250, min_value_out=0, offset_in=0, scale_in=1, out_data_type=np.ubyte):
    z = ma.masked_greater(in_arr, max_value_in)     #mask values greater than max
    z = ma.masked_less(z, min_value_in)  # mask values less than min

    #create physical values
    physical_value = offset_in + z * scale_in   #rescale values
    #adapt spectral the S2 values to PROBAV
    spectral_addapted_array = offset_out + scale_out * physical_value

    #do scaling again
    out_array = (spectral_addapted_array - offset_in) / scale_in

    #Keep track of saturation
    out_array = ma.where( out_array > max_value_out, max_value_out, out_array)
    out_array = ma.where( out_array < min_value_out, min_value_out, out_array)

    if out_data_type in [np.ubyte, np.int16]:
        not_rounded = out_array
        out_array = np.floor(out_array + 0.5).astype(out_data_type)    # use rounding half up method
    out_array = out_array.filled(in_arr)        #fill in original masked values again 251->255
    return out_array

def doRescale(roi, in_file, out_file, new_offset, new_scale, producttype):
    if not producttype in ['ALB','NDV']:
        raise Exception("producttype should be ALB|NDV not %s" % producttype)

    scale = lv2_settings[producttype +'_SCALE'],
    offset = lv2_settings[producttype +'_OFFSET'],
    description = lv2_settings[producttype +'_DESCRIPTION'],
    units = lv2_settings[producttype +'_UNITS'],
    flags = lv2_settings[producttype +'_FLAGS']
    out_nodata = lv2_settings[producttype +'_NODATA']

    os.environ['GDAL_CACHEMAX'] = lv2_settings['GDAL_CACHEMAX']
    #load the source profile which we can reuse for the destination
    with rasterio.open(in_file, 'r', sharing=False ) as src_sr:
        sr_profile = src_sr.profile
        #set output
        profile = sr_profile
        profile.update(
            dtype=rasterio.uint8,
            count=1,
            compress='lzw',
            BIGTIFF='YES'
        )

    if os.path.exists(out_file):
        try:
            os.remove(out_file)
        except:
            print("Warning unable to delete %s" % out_file)
    # open all files for blockprocessing
    with rasterio.open(out_file, 'w', sharing=False, **profile) as dst:
        with rasterio.open(in_file, 'r', sharing=False) as src:
            for ji, window in src.block_windows(1):
                in_array = src.read(1, window=window)
                # do the albedo processing
                rescaled_arr = rescale_S2_Proba(in_array, new_offset, new_scale, max_value_in=250, min_value_in=0,
                                                max_value_out=250,
                                                min_value_out=0, offset_in=offset, scale_in=scale,
                                                out_data_type=np.ubyte)
                # write result to disk
                dst.write(rescaled_arr.astype(rasterio.uint8), 1, window=window)
        dst.nodata = out_nodata

    #update tif metadata
    out_file = tif_update_metadata(out_file,
                                    scale=scale,
                                    offset=offset,
                                    description=description,
                                    units=units,
                                    flags=flags
                                    )
    print("Finished successfully with writing %s" % out_file)

    return out_file



# def FAPAR(NDVI_in):
#     S2_COEFS = [-0.003722892506502186, -0.3868060534807155,  7.055361461275105, -12.43563702856161,  7.109692968647273]    #POLYNOMIAL COEFFICIENTS
#     #PROBAV_COEFS =[]
#     #MODIS_COEFS=[]
#     src = rasterio.open(NDVI_in)
#     profile = src.profile.copy()
#     arr_in = src.read(1)
#     src.close()
#
# def outDir(out_dekad_dir):
#     if not os.path.exists(out_dekad_dir):
#         os.mkdirs(out_dekad_dir)
#     return out_dekad_dir

def start_postproc( mosDir, outputDir, roi, dekad, prev_qua_dek_img=None):
    #in NRT, the smoother wil generate 6 mosaicks back in time, all of these should be handled
    #dekad_list = getHistoricalDekadList(current_dekad, nrOfHistoricalDekads)
    file_prefix = 'L2' if roi == 'AFR' else 'L2_%s' % roi

    #create folders if they do not exist
    alb_dir = join(outputDir, 'alb', 'smo')
    ndv_dir = join(outputDir, 'ndv', 'smo')
    fpr_dir = join(outputDir, 'fpr', 'smo')
    qua_dir = join(outputDir,  'ndv','qua')
    current_dek_img = join(qua_dir, '%s_QUAL_%s.tif') % (file_prefix, dekad)
    for out_folder in [alb_dir, ndv_dir, fpr_dir, qua_dir]:
        if not os.path.exists(out_folder):
            try:
                os.makedirs(out_folder)
            except:
                raise Exception('Unable to create output folder %s' % out_folder)

    #validate inputs
    full_in_alb_path = join(mosDir, 'composite_ALB_%s.tif') % (dekad)
    full_in_ndvi_path = join(mosDir,  'composite_NDV_%s.tif') % ( dekad)
    current_qua_day_img = join(mosDir, 'composite_QUA_%s.tif') % (dekad)
    files_2_validate = [full_in_alb_path, full_in_ndvi_path, current_qua_day_img]

    if prev_qua_dek_img == None:
        #if no image is given, build it yourselve, if the file does not exist at the end, use flag values
        prev_dekad = previousDekad(dekad)
        prev_qua_dek_img = current_dek_img.replace(str(dekad), str(prev_dekad))
        #if not exist, use flag values
        if not os.path.exists(prev_qua_dek_img):
            prev_qua_dek_img = None

    if prev_qua_dek_img != None:
        files_2_validate.append(prev_qua_dek_img)

    validate_inputs(files_2_validate)




    #### SPECTRAL ADAPTATIONS (remove spectral difference between PROBA and S2) ####
    #Albedo
    full_out_path = join(alb_dir, '%s_ALB_%s.tif') % (file_prefix, dekad)
    alb_file = doRescale(roi, full_in_alb_path, full_out_path, 0.01125655723807689, 0.9778930371306196, 'ALB' )
    #Ndvi
    full_out_path = join(ndv_dir, '%s_NDVI_%s.tif') % (file_prefix, dekad)
    ndvi_file = doRescale(roi, full_in_ndvi_path, full_out_path, 0.02196874232754293, 1.016066238343899, 'NDV')
    #create quicklook
    ql_width = lv2_settings['ROIS'][roi]['ql_width']
    ql_height = lv2_settings['ROIS'][roi]['ql_height']
    create_quicklook(ndvi_file, 'ndvi.txt', height=ql_height, width=ql_width)

    #FAPAR
    #fapar is calculated from the original ndvi file, not the spectral corrected one
    full_in_fapar_path = join(mosDir, 'composite_NDV_%s.tif') % (dekad)
    full_out_path = join(fpr_dir, '%s_FPR_%s.tif') % (file_prefix, dekad)
    fapar_file = start_fapar(full_in_fapar_path, lv2_settings['S2_POLYNOMIAL_COEFFICIENTS'], full_out_path, NDVImin=None)

    #QUAadapt
    quaDEK_file = start_qua_adapt(current_qua_day_img, current_dek_img, prev_qua_dek_img, True )

    print('Byebye')


def main():
    args = parse_arguments()
    start_postproc(args.input_dir, args.output_dir, args.roi, args.dekad, args.prev_qua_dek_img)


    # testarray
    # test_arr = np.array([[0, 100, 250], [1, 251, 254]], dtype='int16')
    # print(test_arr.shape)


if __name__ == "__main__":
    main()