import argparse
from pyspark import SparkContext
from pyspark import SparkConf
import subprocess
import os
from os.path import join
import traceback
import rasterio
import numpy as np

from frame_s2.lv1_import_and_mask import start_import_and_mask_lv1
from frame_s2.smooth15 import start_smooth15
from frame_s2.fapar import start_fapar
from frame_s2.config.FRAME_probav_settings  import proba_settings
from frame_s2.postproc_helper import tif_update_metadata, dateToDekad, getHistoricalDekadList, dekadToStartDate
from frame_s2.albedo import calc_and_mask_albedo

#start docker containter
# docker run -v /data/MEP_Projects/Frame:/data/projects/Frame:ro -v ~/Private/Frame/output:/data/output:rw -it frame_smooth
#frame_s2smooth --output-end 2021-03-10 --input-end 2021-03-10 --input-dekads 36 --output-dekads 6 --output-template s2t_{prd}_{dekad} --mask /data/projects/Frame/nrt/S2/REF/S2_SRI_mask.tif /data/projects/Frame//nrt/S2/TILES/44PMQ/TIF/ /data/output/TILES/44PMQ/T10/2107/ --dump-metadata /data/output/logs/44PMQ_2107.log
#

def parse_arguments():

    parser = argparse.ArgumentParser(prog='postprocess_all_probav_lv2.py')

    parser.add_argument("--input_folder_lst", type=str,
                        help="list of proba folders(fullpath, separated by newline) to process. e.g /vitodata/frame_5/lv2/syn/pbv/20150101",
                        required=True)
    parser.add_argument("--root_dir", type=str,
                        help="proba root folder, with subdirs=[alb,ndv,fpr,...]",
                        required=True)
    parser.add_argument("--projwin", type=str,
                        help="-projwin ulx uly lrx lry, used by gdal_translate",
                        required=False,
                        default=None)
    # parser.add_argument("--sensor", type=str,
    #                     help = "sensor, can be MODIS(default) or PROBA",
    #                     default='MODIS')
    return parser.parse_args()




def run_command(cmd):
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, stderr = p.communicate()
    err_nr = p.returncode
    return str(stdout), str(stderr), err_nr

def import_proba(preprocessing_dir, outDir, dekad, dekad_date, projwin):
    print("Converting img to tif ....")
    if projwin is not None:
        print ("Clipping file to %s" % projwin)
        projwin = "-projwin %s " %projwin
    else:
        projwin = ""

    for band, bandproperties in proba_settings['proba_bands'].items():
        img_fname = join(preprocessing_dir, dekad_date, 'pt%s%s.img') % (dekad, band)
        tif_fname = join(outDir, dekad_date, 'pt%s%s.tif') % (dekad, band)
        tif_dir = os.path.dirname(tif_fname)
        if not os.path.exists(tif_dir):
            try:
                os.mkdir(tif_dir)
            except:
                raise Exception("Unable to create %s" % tif_dir)
        cmd = "gdal_translate %s -co 'BLOCKXSIZE=256' -co 'BLOCKYSIZE=256' -co 'TILED=YES' -co 'COMPRESS=LZW' %s %s" % (
        projwin, img_fname, tif_fname)
        #cmd = "echo No gdal_translate in testing mode"
        stdout, stderr, err_nr = run_command(cmd)
        if err_nr:
            print(traceback.print_exc())
            print(stdout)
            print(stderr)
            raise Exception("Converting img to tif failed for file %s to " % (img_fname,tif_fname))

    print("  Finished successfully with writing tifs to folder %s" % tif_dir)
    return tif_dir

def mask_data(in_array, sm_array, mask, new_mask_val=255, pixel_convert_lst=[]):
    #proba SM-map : Status bits: MSB 7-4=BLUE/RED/NIR/SWIR OK; 3=land; 2=snow; 1=cloud; LSB 0=[shadow]
    OF_mis = 251    # data error in RED / NIR(~ SM)
    OF_clo = 252    # cloud or cloud shadow(~ SM)
    OF_sno = 253    # snow / ice(~ SM)
    OF_sea = 254    # sea(~ Land Mask)
    OF_out = 255;   # unknown(no Land Mask & & NDVI = 0) == > boreal zones in winter(land and water  mixed)
    if not in_array.shape == mask.shape:
        raise Exception('Shapes are different for mask and in_array')

    in_array = np.where(in_array > 250, OF_mis, in_array)  # set all flagged in to OF_mis
    in_array = np.where(np.bitwise_and(sm_array, 3) != 0, OF_clo, in_array)      #252=Cloud/shadow according to SM       # 3  : 00000011
    in_array = np.where(np.bitwise_and(sm_array, 4) == 4, OF_sno, in_array)      #253=Snow/Ice according to SM           # 4  : 00000100
    in_array = np.where(np.bitwise_and(sm_array, 96) != 96, OF_mis, in_array)    #Error in RED or NIR according to SM    # 96 : 01100000
    in_array = np.where(sm_array == 255, OF_mis, in_array)
    in_array = np.where(sm_array == 0, OF_mis, in_array)

    in_array = np.where(mask < 1, OF_sea, in_array)  # set all sea pixels to OF_sea

    return in_array


def mask_proba_ndvi(in_ndvi_file, in_sm_file, out_ndvi_file, mask_file):
    with rasterio.open(in_ndvi_file, 'r+', sharing=False) as src:
        profile = src.profile

    # set output
    profile.update(
        dtype=rasterio.uint8,
        count=1,
        compress='lzw',
        BIGTIFF='YES')

    with rasterio.open(out_ndvi_file, 'w+', **profile, sharing=False) as dst:
        with rasterio.open(in_ndvi_file, 'r+', sharing=False) as src:
            with rasterio.open(in_sm_file, 'r+', sharing=False) as sm_src:
                with rasterio.open(mask_file, 'r+', sharing=False) as msk:
                    print("masking ndvi ....")
                    for ji, window in dst.block_windows(1):
                        #print ('ji : %s, %s' % (ji[0], ji[1]))
                        ndvi_in = src.read(1, window=window).astype(np.ubyte)
                        sm_in = sm_src.read(1, window=window).astype(np.ubyte)
                        msk_array = msk.read(1, window=window).astype(np.ubyte)
                        ndvi_out = mask_data(ndvi_in, sm_in, msk_array, pixel_convert_lst=[(254, 251)] )
                        dst.write(ndvi_out, 1, window=window)

    #update tif metadata
    ndvi_out = tif_update_metadata(out_ndvi_file,
                                    scale=proba_settings['NDV_SCALE'],
                                    offset=proba_settings['NDV_OFFSET'],
                                    description=proba_settings['NDV_DESCRIPTION'],
                                    units=proba_settings['NDV_UNITS'],
                                    flags=proba_settings['NDV_FLAGS']
                                    )
    print("  Finished successfully with writing %s" % ndvi_out)


def start_import_and_mask_proba( root_dir, mask_file, in_folder, normalisation, projwin):
    #in_folder will be the subdir in format YYYYMMDD where all modis files are located
    preprocessing_dir = os.path.dirname(in_folder.rstrip('/'))  #yyyymmdd should be skipped here
    dekad_date = os.path.basename(in_folder.rstrip('/'))
    dekad = dateToDekad(dekad_date)
    #set folders if they do not exist
    #preprocessing_dir = join(root_dir, 'syn/mod')
    alb_dir = join(root_dir, 'alb', 'msk')
    ndv_msk_dir = join(root_dir, 'ndv', 'msk')
    tmp_dir = join(root_dir, 'tmp')
    if not os.path.exists(tmp_dir):
         os.mkdir(tmp_dir)

    # if not os.path.exists(alb_dir):
    #     os.mkdir(alb_dir)

    # if not os.path.exists(ndv_dir):
    #     os.mkdir(ndv_dir)


    ##1) import
    import_dir = import_proba(preprocessing_dir, tmp_dir, dekad, dekad_date, projwin)

    ##2) Masking NDVI
    in_ndvi_file = join(import_dir, 'pt%s5.tif') % dekad
    in_sm_file = join(import_dir, 'pt%s0.tif') % dekad
    out_ndvi_file = join(ndv_msk_dir, 'L2_NDVI_%s.tif') % dekad
    mask_proba_ndvi(in_ndvi_file, in_sm_file, out_ndvi_file, mask_file)

    # 3) Calculate ALBEDO
    mask_file = out_ndvi_file
    albedo_file = calc_and_mask_albedo(import_dir, alb_dir, mask_file, dekad, NORM=normalisation, sensor='PROBA')

    print('Done masking NDVI & ALBEDO')


def process_file( folder_name, root_dir, projwin ):
    os.environ['GDAL_CACHEMAX'] = proba_settings['GDAL_CACHEMAX']
    #output_dir = /vitodata/Frame/Lv1/
    folder_name = folder_name.strip() #in format YYYYMMDD
    # we use input start date for historical processing and input-dekads for nrt processing
    print ("Importing and masking = %s" % folder_name)
    start_import_and_mask_proba(root_dir, proba_settings['LS_MASK_FILE'], folder_name, proba_settings['ALB_NORMALISATION'], projwin)

    # ##SMOOTHING
    # end_date_yyyymmdd = os.path.basename(folder_name)
    # current_dekad = dateToDekad(end_date_yyyymmdd)
    # start_in_dekad = getHistoricalDekadList(current_dekad, 36)[35]
    # start_out_dekad = getHistoricalDekadList(current_dekad, 6)[5]
    # start_in_date_yyyymmdd = dekadToStartDate(start_in_dekad)
    # start_out_date_yyyymmdd = dekadToStartDate(start_out_dekad)
    # # Smoothing NDVI
    # #smooth_main(
    # # ndvi, in_file_prefix, in_file_suffix, start_in, end_in, max_missing, hdr_in,
    # # out_file_prefix, out_file_suffix, start_out, end_out,
    # # ymin, ymaxdip,nvimaxmix, QUA, ydelt)
    # #NDVI + QUA
    # start_smooth15( 1,   join(root_dir, 'ndv', 'msk', 'L2_NDVI_'), '', start_in_date_yyyymmdd,  end_date_yyyymmdd, 0, '0.004 -0.08 0 250',
    #              join(root_dir, 'ndv', 'smo', 'L2_NDVI_'), '', start_out_date_yyyymmdd, end_date_yyyymmdd,
    #              0, 0.15, 99, 1, 0.025)
    #
    # #ALBEDO #TODO the 200 max value kan ook 140 zijn
    # start_smooth15(0,   join(root_dir, 'alb', 'msk', 'L2_ALB_'), '', start_in_date_yyyymmdd,  end_date_yyyymmdd, 0, '0.005 0 0 200',
    #              join(root_dir, 'alb', 'smo', 'L2_ALB_'), '',  start_out_date_yyyymmdd, end_date_yyyymmdd,
    #              0, 0.00, 99, 1, 0.025)
    #
    # #FAPAR generation
    # # fapar
    # in_dir = join(root_dir, 'ndv', 'smo')  # previous nrt versions(until currentdekad-5) will be overwritten
    # for dekad in getHistoricalDekadList(current_dekad, 6):
    #     in_file = join(in_dir, 'L2_NDVI_%s.tif' % dekad)
    #     fapar_file = join(root_dir, 'fpr', 'smo', 'L2_FPR_%s.tif' % dekad)
    #     start_fapar(in_file, proba_settings['FPR_POLYNOMIAL_COEFFICIENTS_PROBAV'], fapar_file)


def main():
    args = parse_arguments()
    print(args)

    sc_conf = SparkConf()
    sc = SparkContext(conf=sc_conf)


    if not os.path.exists(args.input_folder_lst):
        raise Exception("The input folder list does not exist : %s" % args.input_folder_lst)
    try:
        with open(args.input_folder_lst) as f:
            in_files = f.readlines()
        #in_files = list(map(lambda fname: fname.replace('"',''), in_files))  # remove " around tiles
        number_of_files = len(in_files)
        if number_of_files > 0:
            #logdir = os.path.join(args.output_dir, 'logs')
            tiles_rdd = sc.parallelize(in_files, len(in_files))
            tiles_rdd.foreach(lambda fname: process_file(fname, args.root_dir, args.projwin))
        else:
            raise Exception("input list is empty %s" % args.input_folder_lst )

    finally:
        sc.stop()


if __name__ == "__main__":
    main()

