# -*- coding: utf-8 -*-
import argparse
import rasterio
import numpy as np
import numpy.ma as ma
import os
from os.path import join

from frame_s2.postproc_helper import validate_inputs, tif_update_metadata, getDekadLst, previousDekad
from frame_s2.config.FRAME_settings import lv1_settings

def parse_arguments():
    parser = argparse.ArgumentParser(prog='qua_adapt_lv1.py',
                                     description="qua_adapt used for lv1. This script will convert "
                                                 "QUA(gap between surrounding good observations!) expressed in days "
                                                 "to QUA (gap since previous good observation) expressed in dekads")
    
    parser.add_argument("--in_qua_prefix", type=str, help="qua_day file prefix(with path)",
                        required=False,
                        default=None)
    parser.add_argument("--out_qua_dek_prefix", type=str, help="qua_dek file prefix(with path)",
                        required=False,
                        default=None)
    parser.add_argument("--start_dekad", type=int, help="start dekad to calculate qua_dek(YYTT)",
                         required=True)
    parser.add_argument("--end_dekad", type=int, help="end dekad to calculate qua_dek (YYTT)",
                        required=True)

    return parser.parse_args()


def main():
    args = parse_arguments()
    start_qua_adapt(args.in_qua_prefix, args.out_qua_dek_prefix, args.start_dekad, args.end_dekad)

def start_qua_adapt( in_qua_prefix, out_qua_dek_prefix, start_dekad, end_dekad):
    #set gdal cache parameter on the workers to reduce memory usage.
    # rasterio uses gdal in the background so limit the GDAL_CACHE
    # this parameter is in bytes. 1GB wil result in 1,2GB of memory usage(tested on local machine)
    os.environ['GDAL_CACHEMAX'] = lv1_settings['GDAL_CACHEMAX']

    #initialize
    hist_gap = 0        #gap beteween current dekad and previous_out_qua_Dek
    infiles_2_validate = []
    dek_list = getDekadLst(start_dekad, end_dekad)
    infiles_2_validate = list(map(lambda dekad: join(in_qua_prefix + dekad + '.tif'), dek_list))
    initial_hist_dekad = previousDekad(start_dekad)   #start with this historical dekad
    initial_hist_file = join(out_qua_dek_prefix + initial_hist_dekad + '.tif')
    infiles_2_validate.append(initial_hist_file)
    # for f in in_qua_day_file_lst:
    #     if not os.path.exists(f):
    #         raise Exception("qua_day image does not exist: %s" % f)


    # #get inital image
    # infiles_2_validate=[quaday_img]
    # if previous_out_qua_dek != None:
    #     try:
    #         hist_dekad = re.search('[0-9]{4}', os.path.basename(previous_out_qua_dek)).group(0)
    #         current_dekad = re.search('[0-9]{4}', os.path.basename(current_dek_img)).group(0)
    #     except:
    #         raise Exception('Unable to get dekad out of filenames [%s,%s]' %(os.path.basename(previous_out_qua_dek),os.path.basename(current_dek_img)))
    #     hist_gap = int(current_dekad) - int(hist_dekad)
    #     #use previous dekad to get initial values.
    #     src_prev = rasterio.open(previous_out_qua_dek)
    #     infiles_2_validate.append(previous_out_qua_dek)

    #validate all inputs
    validate_inputs(infiles_2_validate)

    #creating profile
    with rasterio.open(initial_hist_file) as src:
        src_profile = src.profile
        # set output
        profile = src_profile
        profile.update(
            dtype=rasterio.uint8,
            count=1,
            compress='lzw',
            BIGTIFF='YES')

    previous_out_qua_dek = initial_hist_file
    for dekad in dek_list:
        #set previous_dekad_file
        if not os.path.exists(previous_out_qua_dek):  # if not exist, use an empty array later on
            previous_out_qua_dek = None
        else:
            src_prev = rasterio.open(previous_out_qua_dek)

        quaday_img = join(in_qua_prefix + dekad + '.tif')
        current_dek_img = join(out_qua_dek_prefix + dekad + '.tif')
        # start processing dekad by dekad, starting with the first
        with rasterio.open(quaday_img) as src_curr:
            with rasterio.open(current_dek_img, 'w', **profile) as dst:
                for ji, window in src_curr.block_windows(1):
                    current_day = src_curr.read(1, window=window)

                    if previous_out_qua_dek != None:
                        prev_dek = src_prev.read(1, window=window)
                    else:
                        prev_dek = np.full(current_day.shape, 251, np.uint8)        #use 251, then prev will be skipped and only the day image is used, if 254, it will be masked

                    #convert gap[days] between surrounding (previous->next) to previous gap [dekads]
                    current_dek = calc_qua_lv1(current_day, prev_dek, keep250=True)
                    dst.write(current_dek, window=window, indexes=1)

                dst.nodata = 255

        if previous_out_qua_dek != None:
            src_prev.close()
        previous_out_qua_dek = current_dek_img #current_dekad file becomes the previous one in the next run

        #update tif metadata
        current_dek_img = tif_update_metadata(current_dek_img,
                                        scale=lv1_settings['QUA_SCALE'],
                                        offset=lv1_settings['QUA_OFFSET'],
                                        description=lv1_settings['QUA_DESCRIPTION'],
                                        units=lv1_settings['QUA_UNITS'],
                                        flags=lv1_settings['QUA_FLAGS']
                                        )
        print("Finished successfully with writing %s" % current_dek_img)
    return current_dek_img

def calc_qua_lv1(current_day, hist, keep250=True):
    # all values > 251 will be flagged to 254
    #basic rule : use previous dek image +1, but reset if there as a valid observation in this dekad

    # convert acceptable observations to good observations
    #msk_255_day = np.where(current_day == 255, 1, 0)
    current_day_masked = ma.masked_greater(current_day, 252)
    if keep250:
        current_day_masked = ma.where(current_day_masked == 250, 0,
                                      current_day_masked)  # special case: if there is a small number of observations in the smoother, it will set this to 250. We will hard code this to 0

    # mask missing land pixels
    prev_dek_masked = ma.masked_greater(hist, 252)  # expressed in dekads
    current_dek_masked = ma.where(current_day_masked == 0, 0, np.minimum(prev_dek_masked + 1, 251)) #solve saturation but keep 251
    #current_dek_masked = ma.where(current_day_masked == 251, 251, current_dek_masked )  #copy bad day_pixels to dek_image
    # prev_dek_masked = ma.where(prev_dek_masked < 251, prev_dek_masked + hist_gap,
    #                            prev_dek_masked)  # do not increase missing land data
    current_dek_masked = ma.where(current_dek_masked < 251, np.minimum(current_dek_masked, 240),
                               current_dek_masked)                                      # solve saturation, but keep 251
    current_day_masked = None

    # from here, all calculations are expressed in dekads
    current_dek = current_dek_masked.filled(fill_value=254)  # reset flagged values
    #current_dek = np.where(msk_255_day == 1, 255, current_dek)  # set 255 pixels again
    #msk_255_day = None
    return current_dek

#
# def calc_qua_using_hist(current_day, hist, hist_gap, keep250=True, S2=True):
#     #all values > 251 will be flagged to 254
#
#     #convert acceptable observations to good observations
#     current_day_masked = ma.masked_greater(current_day, 252)
#     if keep250:
#         current_day_masked = ma.where(current_day_masked == 250, 0, current_day_masked) #special case: if there is a small number of observations in the smoother, it will set this to 250. We will hard code this to 0
#
#     #mask missing land pixels
#     #current_day_masked_values = ma.masked_less(current_day, 251)                #expressed in days
#     prev_dek_masked = ma.masked_greater(hist, 252)                              #expressed in dekads
#     prev_dek_masked = ma.where(prev_dek_masked < 251, prev_dek_masked + hist_gap, prev_dek_masked) # do not increase missing land data
#     prev_dek_masked = ma.where(prev_dek_masked < 251 ,np.minimum(prev_dek_masked, 240), prev_dek_masked)                ## solve saturation, but keep 251
#
#     dif = 0 if S2 else 1  #qua_day for smooth15 is never 10(1dekad). At least these are 20/21 days
#     current_dek_masked = ma.where(current_day_masked < 240, np.maximum((current_day_masked / 10 - dif), 0).astype(np.ubyte) ,  current_day_masked)        # if not S2, do minus 1, but set to 0 if results become negative
#     current_day_masked = None
#
#     #from here, all calculations are expressed in dekads
#
#     current_dek_masked = ma.minimum(current_dek_masked, prev_dek_masked)
#     #current_dek_masked = ma.where(current_dek_masked < hist_gap, current_dek_masked, prev_dek_masked + hist_gap)   #use prev value if current image has no matches
#     current_dek_masked = ma.where(current_dek_masked < 251, np.minimum(current_dek_masked, 240), current_dek_masked)   # solve saturation, but keep 251
#     current_dek = current_dek_masked.filled(fill_value=254)    # reset flagged values
#     return current_dek
#
# def calc_qua_herman_like(current_day, prev):
#     current_day_masked = ma.masked_greater(current_day, 250)
#     current_day_masked = ma.where(current_day_masked == 250, 0, current_day_masked) #special case: if there is a small number of observations in the smoother, it will set this to 250. We will hard code this to 0
#     current_day_masked_values = ma.masked_less(current_day, 251)
#     prev_masked = ma.masked_greater(prev, 251)
#     prev_masked = ma.where(prev_masked == 250, 0, prev_masked)
#
#     current_day_masked = (current_day_masked / 10).astype(np.ubyte)  # from days to dekads
#
#     #use previous result as
#     current_day_masked = ma.where(current_day_masked == 0, 0, prev_masked + 1)
#     current_day_masked = ma.where(current_day_masked >= 240, 240,
#                                   current_day_masked)  # solve saturation
#     current_day = current_day_masked.filled(fill_value=0) + current_day_masked_values.filled(
#         fill_value=0)  # rebuilding the array with nodata values filled in
#     # current_day = current_day_masked.filled(fill_value=flag_out)
#     return  current_day
#
# def calc_qua(current_day):
#     current_day_masked = ma.masked_greater(current_day, 250)
#     current_day_masked = ma.where(current_day_masked == 250, 0, current_day_masked )   #special case: if there is a small number of observations in the smoother, it will set this to 250. We will hard code this to 0
#     current_day_masked_values = ma.masked_less(current_day, 251)
#     current_day_masked = (current_day_masked / 10).astype(np.ubyte)  # from days to dekads
#
#     current_day_masked = ma.where(current_day_masked >= 240, 240,
#                                   current_day_masked)  # solve saturation
#     current_day = current_day_masked.filled(fill_value=0) + current_day_masked_values.filled(
#                     fill_value=0)  # rebuilding the array with nodata values filled in
#                                     # current_day = current_day_masked.filled(fill_value=flag_out)
#     return  current_day
if __name__ == "__main__":
    main()